import hashlib
import json
import logging
from typing import Any

from omegaconf import DictConfig, OmegaConf

_logger = logging.getLogger(__name__)


def hash_dict(data: dict[str, Any]) -> str:
    """Return a hash of a dictionary."""
    string = json.dumps(data, indent=True, sort_keys=True, ensure_ascii=True)
    return hashlib.sha256(string.encode()).hexdigest()


def hash_config(cfg: DictConfig) -> str:
    """Compute a unique identifier that summarizes the provided config."""
    if isinstance(cfg, DictConfig):
        config_dict: dict = OmegaConf.to_container(cfg, resolve=True)  # type: ignore
    else:
        config_dict = dict(cfg)
    return hash_dict(config_dict)


def display_config(config: DictConfig = None, logger=None, **kwargs) -> int:
    """Display the configuration."""
    try:
        config_uuid = hash_config(config)
    except Exception as e:
        if logger:
            logger.error(f"Error hashing config {config}: {e}")
        config_uuid = None
    if logger:
        logger.info(f"Loaded config with uuid {config_uuid}")
    if kwargs:
        logger.warning("The configuration is a dictionary, not a DictConfig object.")
        if config is not None:
            kwargs["config"] = config
        to_print = json.dumps(kwargs, indent=4)
        if logger:
            logger.info(f"Config values:\n{to_print}")
        return 0

    try:
        to_print = json.dumps(OmegaConf.to_container(config, resolve=True), indent=4)
        if logger:
            logger.info(f"Config values:\n{to_print}")
    except Exception as e:
        if logger:
            logger.error(f"Error resolving config: {e}")
        try:
            to_print = json.dumps(OmegaConf.to_container(config), indent=4)
            if logger:
                logger.info(f"Config values:\n{to_print}")
        except Exception as e:
            if logger:
                logger.error(f"Error printing config: {e}")
                logger.info(f"Config values:\n{config}")
        return 1

    return 0



from typing import Optional, Callable
import os

import click
from click import Context, Option


def get_no_terminal_efects() -> bool:
    """Get the no terminal effects environment variable."""
    val = os.environ.get(
        "OMEGACLICK_NO_TERMINAL_EFFECTS", os.environ.get("OMEGACLICK_NO_TERMINAL_EFFECTS")
    )
    if val is None:
        return False
    return val.lower() in {"true", "1", "yes"}


NO_TERMINAL_EFFECTS = get_no_terminal_efects()


def display_terminal_effect(value):
    """Display the terminal effect."""
    # from terminaltexteffects.effects.effect_random_sequence import RandomSequence
    from terminaltexteffects.effects.effect_print import Print  # noqa: PLC0415
    from terminaltexteffects.utils.graphics import Color  # noqa: PLC0415

    effect = Print(value)
    # effect.effect_config.speed = 0.025
    effect.effect_config.print_speed = 5
    effect.effect_config.print_head_return_speed = 3
    effect.effect_config.final_gradient_stops = (Color("00ffae"), Color("00D1FF"), Color("FFFFFF"))
    with effect.terminal_output() as terminal:
        for frame in effect:
            terminal.print(frame)


_TERMINAL_EFFECT = display_terminal_effect


def set_terminal_effect(value: Callable | None = None):
    """Set the terminal effect."""
    global _TERMINAL_EFFECT  # noqa: PLW0603
    if value is None:
        value = display_terminal_effect
    _TERMINAL_EFFECT = value


def parse_args(self, ctx: Context, args: list[str]) -> list[str]:
    """Display the help message when no arguments are provided."""
    if not args and self.no_args_is_help and not ctx.resilient_parsing:
        _TERMINAL_EFFECT(ctx.get_help())
        ctx.exit()

    rest = super(click.core.MultiCommand, self).parse_args(ctx, args)

    if self.chain:
        ctx.protected_args = rest
        ctx.args = []
    elif rest:
        ctx.protected_args, ctx.args = rest[:1], rest[1:]

    return ctx.args


def get_help_option(self, ctx: Context) -> Optional["Option"]:
    """Return the help option object."""
    from gettext import gettext  # noqa: PLC0415

    help_options = self.get_help_option_names(ctx)

    if not help_options or not self.add_help_option:
        return None

    def show_help(ctx: Context, param: "click.Parameter", value: str) -> None:  # noqa: ARG001
        if value and not ctx.resilient_parsing:
            display_terminal_effect(ctx.get_help())
            ctx.exit()

    return Option(
        help_options,
        is_flag=True,
        is_eager=True,
        expose_value=False,
        callback=show_help,
        help=gettext("Show this message and exit."),
    )


if not NO_TERMINAL_EFFECTS:
    set_terminal_effect(display_terminal_effect)
    click.core.Command.get_help_option = get_help_option
    click.core.MultiCommand.parse_args = parse_args



import click


hydra_args_argument = click.argument("hydra_args", nargs=-1, type=click.UNPROCESSED)


# Information about hydra
hydra_help_option = click.option(
    "--hydra-help",
    "hydra_help",
    help="Shows Hydra specific flags",
    is_flag=True,
    default=False,
)

version_option = click.option(
    "--version",
    "version",
    help="Show Hydra's version and exit",
    is_flag=True,
    default=False,
)

# Debugging assistance
show_config_option = click.option(
    "--cfg",
    "-c",
    "show_config",
    help="Show config instead of running. Takes as parameter one of job, hydra or all",
    type=click.Choice(["job", "hydra", "all"]),
    default=None,
)
resolve_option = click.option(
    "--resolve",
    "resolve_",
    help=(
        "Used in conjunction with the --cfg flag; "
        "resolve interpolations in the config before printing it"
    ),
    is_flag=True,
    default=False,
)
package_option = click.option(
    "--package",
    "-p",
    "package",
    help="Used in conjunction with --cfg to select a specific config package to show",
    type=click.STRING,
)
info_option = click.option(
    "--info",
    "-i",
    "info",
    help=(
        "Print Hydra information. This includes installed plugins, "
        "Config Search Path, Defaults List, generated config and more"
    ),
    is_flag=True,
    default=False,
)
# Running Hydra applications:
run_option = click.option(
    "--run",
    "run",
    help="Run is the default mode and is not normally needed.",
    is_flag=True,
    default=False,
)
multirun_option = click.option(
    "--multirun",
    "-m",
    "multirun",
    help="Run multiple jobs with the configured launcher and sweeper",
    is_flag=True,
    default=False,
)
config_path_option = click.option(
    "--config-path",
    "-cp",
    "config_path",
    help=(
        "Overrides the config_path specified in hydra.main(). "
        "The config_path is absolute or relative to the Python file declaring @hydra.main()."
    ),
    default=None,
    type=click.Path(
        exists=False,
        file_okay=False,
        dir_okay=True,
        readable=True,
        resolve_path=True,
        allow_dash=False,
        path_type=None,
    ),
)
config_name_option = click.option(
    "--config-name",
    "-cn",
    "config_name",
    help="Overrides the config_name specified in hydra.main()",
    default=None,
    type=click.STRING,
)
config_dir_option = click.option(
    "--config-dir",
    "-cd",
    "config_dir",
    help=(
        "Adds an additional config directory to the config search path. "
        "This is useful for installed apps that want to allow their users"
        " to provide additional configs"
    ),
    default=None,
    type=click.Path(
        exists=False,
        file_okay=False,
        dir_okay=True,
        readable=True,
        resolve_path=True,
        allow_dash=False,
        path_type=None,
    ),
)
shell_completion_option = click.option(
    "--shell-completion",
    "-sc",
    "shell_completion",
    help="Install or Uninstall shell tab completion",
    is_flag=True,
    default=False,
)
# Unused stuff
parallel_option = click.option(
    "--parallel",
    "-p",
    "parallel",
    help="Run each shard and configuration sweep in parallel",
    is_flag=True,
    default=False,
)
num_shards_option = click.option(
    "--num-shards",
    "-n",
    help="Number of shards to split the data processing",
    default=0,
)

file_option = click.option(
    "--file",
    "-f",
    "file",
    help="Path to the configuration file",
    default=None,
    type=click.Path(
        exists=False,
        file_okay=True,
        dir_okay=False,
        readable=True,
        resolve_path=True,
        allow_dash=False,
        path_type=None,
    ),
)



import sys

import click
import flogging

from hydraclick import hydra_command


@click.group()
def cli():
    """Run command line interface for hydraclick."""
    flogging.setup(allow_trailing_dot=True)


@cli.command(short_help="test_stuff.")
@hydra_command()
def nothing(args, **kwargs):  # noqa: ARG001
    """Test function that does nothing."""


if __name__ == "__main__":
    sys.exit(cli())



import functools
import logging
import sys
from pathlib import Path
from typing import Callable, Any

import hydra
from omegaconf import DictConfig, OmegaConf
from unittest.mock import patch
import flogging

from hydraclick.display_config import display_config
from hydraclick.options import (
    hydra_args_argument,
    hydra_help_option,
    version_option,
    show_config_option,
    resolve_option,
    package_option,
    info_option,
    run_option,
    multirun_option,
    config_path_option,
    config_dir_option,
    config_name_option,
    shell_completion_option,
)

_logger = logging.getLogger(__name__)


def wrap_kwargs_and_config(
    function: Callable,
    run_mode: str = "config",  # "config" | "kwargs"
    print_config: bool = True,
    preprocess_config: Callable[[DictConfig], DictConfig] | None = None,
    resolve: bool = True,
):
    """Wrap a function to run as a hydra command."""

    @functools.wraps(function)
    def wrapper(config):
        if preprocess_config:
            config = preprocess_config(config)
        if isinstance(config, DictConfig) and resolve:
            try:
                OmegaConf.resolve(config)
            except Exception as e:
                display_config(config, logger=_logger)
                raise e
        if print_config:
            display_config(config, logger=_logger)
        if run_mode == "config":
            return function(config)
        return function(**config)

    return wrapper


def build_hydra_args(
    hydra_help: bool,
    version: bool,
    show_config: str,
    resolve: bool,
    package: str,
    info: bool,
    run: bool,
    multirun: bool,
    config_path: str,
    config_name: str,
    config_dir: str,
    shell_completion: bool,
    hydra_args: tuple[str, ...] | None = None,
) -> tuple[str, ...]:
    """Compose the arguments for the hydra command."""
    _logger.debug(f"Hydra args: {hydra_args}")
    hydra_args_ = []
    if hydra_help:
        hydra_args_.append("--hydra-help")
    if version:
        hydra_args_.append("--version")
    if show_config:
        hydra_args_.extend(("--cfg", show_config))
    if resolve:
        hydra_args_.append("--resolve")
    if package:
        hydra_args_.extend(("--package", f"{package}"))
    if info:
        hydra_args_.append("--info")
    if run:
        hydra_args_.append("--run")
    if multirun:
        hydra_args_.append("--multirun")
    if config_path:
        hydra_args_.extend(("--config-path", f"{config_path}"))
    if config_name:
        hydra_args_.extend(("--config-name", f"{config_name}"))
    if config_dir:
        hydra_args_.extend(("--config-dir", f"{config_dir}"))
    if shell_completion:
        hydra_args_.append("--shell-completion")
    _logger.debug(f"Hydra args after composition: {hydra_args}")
    return (*hydra_args_, *hydra_args)


def get_default_dir() -> str:
    """Get the default directory for the hydra config."""
    curr_dir = Path().cwd()
    # if there is a `config` folder inside curr_dir, return its path, otherwise return curr_dir
    return str(curr_dir / "config") if (curr_dir / "config").exists() else str(curr_dir)


def run_hydra(function: Callable, hydra_args: tuple[str, ...]) -> None:
    """Run a function as a hydra app."""

    @hydra.main(config_path=get_default_dir(), config_name="config", version_base=None)
    @functools.wraps(function)
    def _run_hydra_function(loaded_config: DictConfig):
        flogging.setup(allow_trailing_dot=True)
        return function(loaded_config)

    with patch("sys.argv", [sys.argv[0], *list(hydra_args)]):
        return _run_hydra_function()


def command_api(
    function: Callable[[DictConfig], Any],
    run_mode="config",
    preprocess_config=None,
    print_config=True,
    resolve=True,
) -> Callable:
    """Implement using click the hydra CLI API."""

    @hydra_args_argument
    @hydra_help_option
    @version_option
    @show_config_option
    @resolve_option
    @package_option
    @info_option
    @run_option
    @multirun_option
    @config_path_option
    @config_name_option
    @config_dir_option
    @shell_completion_option
    @functools.wraps(function)
    def click_compatible(
        hydra_help: bool,
        version: bool,
        show_config: str,
        resolve_: bool,
        package: str,
        info: bool,
        run: bool,
        multirun: bool,
        config_path: str,
        config_name: str,
        config_dir: str,
        shell_completion: bool,
        hydra_args: tuple[str, ...] | None = None,
    ):
        nonlocal print_config, run_mode, preprocess_config, resolve
        if show_config:
            print_config = False
        true_func = wrap_kwargs_and_config(
            function, run_mode, print_config, preprocess_config, resolve
        )
        hydra_args = build_hydra_args(
            hydra_help,
            version,
            show_config,
            resolve_,
            package,
            info,
            run,
            multirun,
            config_path,
            config_name,
            config_dir,
            shell_completion,
            hydra_args,
        )
        return run_hydra(true_func, hydra_args)

    return click_compatible


def hydra_command(
    run_mode: str = "config",  # "config" | "kwargs"
    print_config: bool = True,
    preprocess_config: Callable[[DictConfig], DictConfig] | None = None,
    resolve: bool = True,
):
    """Wrap a function so it can run as a hydra command."""

    def decorator(function: Callable):
        return command_api(
            function,
            run_mode=run_mode,
            print_config=print_config,
            preprocess_config=preprocess_config,
            resolve=resolve,
        )

    return decorator



from functools import partial
from pathlib import Path
import logging
import sys
from typing import Callable, Any

import hydra
from omegaconf import DictConfig
from unittest.mock import patch
import flogging

from hydraclick.display_config import display_config

_logger = logging.getLogger(__name__)


def get_hydra_configs(
    config_file: Path | str,
    hydra_args: tuple[str, ...],
) -> list[DictConfig]:
    """Load the necessary configuration for running click commands from a config.yaml file.

    Args:
        config_file: Path to the target config.yaml file. If None, the default \
            config file is loaded.
        hydra_args: Arguments passed to hydra for composing the project configuration.

    Returns:
        configs: list of all the resolved configurations specified by the command arguments.

    """
    config_file = Path(config_file)
    configs = []
    if config_file.exists() and config_file.is_file():
        _logger.info("Loading config file from %s", config_file)
        hydra_args = ["--config-dir", str(config_file.parent), *list(hydra_args)]  # type: ignore

        @hydra.main(
            config_path=str(config_file.parent), config_name=config_file.stem, version_base=None
        )
        def load_config(loaded_config: DictConfig):
            flogging.setup(allow_trailing_dot=True)
            nonlocal configs
            configs.append(loaded_config)

        with patch("sys.argv", [sys.argv[0], *list(hydra_args)]):
            load_config()

    else:
        _logger.error(f"Invalid config file path provided: {config_file}")
        msg = f"Invalid config file: {config_file}"
        raise ValueError(msg)
    return configs


def _run_sequential(
    function: Callable[[DictConfig], Any],
    configs: list[DictConfig],
    num_shards: int = 0,
) -> int:
    """Run the sweep sequentially."""
    for config in configs:
        if num_shards == 0:
            function(config)
            continue
        if num_shards > 0:
            config["num_shards"] = num_shards
        num_shards = config.get("num_shards", 1)
        for shard_ix in range(num_shards):
            config["shard_ix"] = shard_ix
            _logger.info("Running shard %d", shard_ix)
            function(config)
    return 0


def _run_parallel(
    function: Callable[[DictConfig], Any],
    configs: list[DictConfig],
    num_shards: int = 0,
) -> int:
    """Run the sweep sequentially."""
    try:
        import ray  # noqa: PLC0415
    except ImportError:
        _logger.error("Ray is not installed. Please install it with `pip install ray`")
        return 1
    _conf = configs[0]
    ray_opts = _conf.get("ray", {})
    ray_opts["ignore_reinit_error"] = True
    _logger.info("Launching ray with parameters: %s", ray_opts)
    ray.init(**ray_opts)
    run_remote = ray.remote(function)
    requests = []
    _logger.info("Ray launched successfully")
    for config in configs:
        if num_shards == 0:
            req_id = run_remote.remote(config)
            requests.append(req_id)
            continue
        if num_shards > 0:
            config["num_shards"] = num_shards
        num_shards = config.get("num_shards", 1)
        for shard_ix in range(num_shards):
            config["shard_ix"] = shard_ix
            req_id = run_remote.remote(config)
            _logger.info("Running shard %d", shard_ix)
            requests.append(req_id)
    ray.get(requests)
    return 0


def run_function(
    function: Callable[[DictConfig], Any],
    config_file: str | Path | None = None,
    hydra_args: tuple[str, ...] | None = None,
    multirun: bool = True,
    parallel: bool = False,
    num_shards: int = 0,
    only_config: bool = False,
) -> int:
    """Run the function."""
    if multirun:
        hydra_args = ["hydra.mode=MULTIRUN", *list(hydra_args)]
    configs = get_hydra_configs(config_file, hydra_args)
    if only_config:
        function = partial(display_config, logger=_logger)

    if parallel:
        _logger.info("Running in parallel mode")
        return _run_parallel(function, configs, num_shards=num_shards)
    _logger.info("Running in sequential mode")
    return _run_sequential(function, configs, num_shards=num_shards)
